import unittest

from keras_preprocessing.text import Tokenizer

from headliner.preprocessing.keras_tokenizer import KerasTokenizer
from headliner.preprocessing.vectorizer import Vectorizer


class TestPreprocessor(unittest.TestCase):

    def test_vectorize(self):
        data = [('a b c', 'd')]
        tokenizer_encoder = KerasTokenizer()
        tokenizer_decoder = KerasTokenizer()
        tokenizer_encoder.fit([data[0][0]])
        tokenizer_decoder.fit([data[0][1]])
        vectorizer = Vectorizer(tokenizer_encoder, tokenizer_decoder, max_output_len=3)
        data_vectorized = [vectorizer(d) for d in data]
        self.assertEqual([([1, 2, 3], [1, 0, 0])], data_vectorized)

        data = [('a b c', 'd d d d')]
        data_vectorized = [vectorizer(d) for d in data]
        self.assertEqual([([1, 2, 3], [1, 1, 1])], data_vectorized)



from typing import Tuple

import tensorflow as tf


class Encoder(tf.keras.Model):

    def __init__(self,
                 embedding_shape: Tuple[int, int],
                 lstm_size=50,
                 embedding_trainable=True,
                 embedding_weights=None) -> None:
        super(Encoder, self).__init__()
        vocab_size, vec_dim = embedding_shape
        weights = None if embedding_weights is None else [embedding_weights]
        self.embedding = tf.keras.layers.Embedding(vocab_size,
                                                   vec_dim,
                                                   weights=weights,
                                                   trainable=embedding_trainable)
        self.lstm = tf.keras.layers.LSTM(lstm_size,
                                         return_sequences=True,
                                         return_state=True)
        self.lstm_size = lstm_size

    def call(self,
             sequence: tf.Tensor,
             states: tf.Tensor) -> Tuple[tf.Tensor, tf.Tensor, tf.Tensor]:
        embed = self.embedding(sequence)
        output, state_h, state_c = self.lstm(embed, initial_state=states)
        return output, state_h, state_c

    def init_states(self, batch_size: int) -> Tuple[tf.Tensor, tf.Tensor]:
        return tf.zeros([batch_size, self.lstm_size]), \
               tf.zeros([batch_size, self.lstm_size])


class LuongAttention(tf.keras.Model):

    def __init__(self, rnn_size):
        super(LuongAttention, self).__init__()
        self.wa = tf.keras.layers.Dense(rnn_size)

    def call(self, decoder_output, encoder_output):
        score = tf.matmul(decoder_output, self.wa(encoder_output), transpose_b=True)
        alignment = tf.nn.softmax(score, axis=2)
        context = tf.matmul(alignment, encoder_output)
        return context, alignment


class Decoder(tf.keras.Model):

    def __init__(self,
                 embedding_shape: Tuple[int, int],
                 lstm_size=50,
                 embedding_trainable=True,
                 embedding_weights=None) -> None:
        super(Decoder, self).__init__()
        self.lstm_size = lstm_size
        vocab_size, vec_dim = embedding_shape
        weights = None if embedding_weights is None else [embedding_weights]
        self.embedding = tf.keras.layers.Embedding(vocab_size, vec_dim,
                                                   weights=weights,
                                                   trainable=embedding_trainable)
        self.lstm_size = lstm_size
        self.attention = LuongAttention(lstm_size)
        self.lstm = tf.keras.layers.LSTM(lstm_size,
                                         return_sequences=True,
                                         return_state=True)
        self.wc = tf.keras.layers.Dense(lstm_size, activation='tanh')
        self.ws = tf.keras.layers.Dense(vocab_size)

    def call(self, sequence, state, encoder_output):
        embed = self.embedding(sequence)
        lstm_out, state_h, state_c = self.lstm(embed, initial_state=state)
        context, alignment = self.attention(lstm_out, encoder_output)
        lstm_out = tf.concat([tf.squeeze(context, 1), tf.squeeze(lstm_out, 1)], 1)
        lstm_out = self.wc(lstm_out)
        logits = self.ws(lstm_out)
        return logits, state_h, state_c, alignment

    def init_states(self, batch_size):
        return (tf.zeros([batch_size, self.lstm_size]),
                tf.zeros([batch_size, self.lstm_size]))




import tensorflow as tf


def masked_crossentropy(targets: tf.Tensor, logits: tf.Tensor) -> tf.Tensor:
    crossentropy = tf.keras.losses.SparseCategoricalCrossentropy(from_logits=True)
    mask = tf.math.logical_not(tf.math.equal(targets, 0))
    mask = tf.cast(mask, dtype=tf.int64)
    loss = crossentropy(targets, logits, sample_weight=mask)
    return loss



from typing import Dict, Union

import numpy as np
from nltk.translate.bleu_score import sentence_bleu

from headliner.evaluation.scorer import Scorer


class BleuScorer(Scorer):
    """
    Provides BLEU score for a model prediction.
    """

    def __init__(self, tokens_to_ignore=None, weights=(0.25, 0.25, 0.25, 0.25)) -> None:
        """
        Initializes the scorer.

        Args:
            tokens_to_ignore: Tokens to be removed before comparing input and output text.
            weights: Custom weights for 1,2,3,4 grams, e.g. (1, 0, 0, 0) will only measure 1-gram overlaps.
        """
        self.tokens_to_exclude = tokens_to_ignore or []
        self.weights = weights

    def __call__(self, prediction: [Dict[str, Union[str, np.array]]]) -> float:
        tokens_predicted = prediction['predicted_text'].split()
        tokens_output = prediction['preprocessed_text'][1].split()
        tokens_predicted_filtered = [t for t in tokens_predicted if t not in self.tokens_to_exclude]
        tokens_output_filtered = [t for t in tokens_output if t not in self.tokens_to_exclude]
        return sentence_bleu([tokens_output_filtered], tokens_predicted_filtered, weights=self.weights)



import abc
from abc import abstractmethod
from typing import List


class Tokenizer(abc.ABC):
    """
    Encodes text to sequences and decodes sequences to text.
    """

    @abstractmethod
    def encode(self, text: str) -> List[int]:
        """
        Encodes a given string into a sequence of indices.

        Args:
            text: Text to encode.

        Returns: Encoded sequence.
        """
        pass

    @abstractmethod
    def decode(self, sequence: List[int]) -> str:
        """
        Decodees a given sequence into a text.

        Args:
            sequence: Sequence to decode.

        Returns: Decoded text.

        """
        pass

    @property
    @abstractmethod
    def vocab_size(self) -> int:
        """
        Size of token vocab.
        """
        pass




import json
from time import time
from typing import Tuple, List

from sklearn.model_selection import train_test_split
from tensorflow_datasets.core.features.text import SubwordTextEncoder

from headliner.model import SummarizerAttention
from headliner.model.summarizer_transformer import SummarizerTransformer
from headliner.preprocessing import Preprocessor, Vectorizer
from headliner.trainer import Trainer


def read_data_json(file_path: str,
                   max_sequence_length: int) -> List[Tuple[str, str]]:
    with open(file_path, 'r', encoding='utf-8') as f:
        data_out = json.load(f)
        return [d for d in zip(data_out['desc'], data_out['heads']) if len(d[0].split(' ')) <= max_sequence_length]


def read_data(file_path: str) -> List[Tuple[str, str]]:
    data_out = []
    with open(file_path, 'r', encoding='utf-8') as f:
        for l in f.readlines():
            x, y = l.strip().split('\t')
            data_out.append((x, y))
        return data_out


if __name__ == '__main__':
    data_raw = read_data_json('/Users/cschaefe/datasets/welt_dedup.json', 2000)
    train_data, val_data = train_test_split(data_raw, test_size=500, shuffle=True, random_state=42)
    """
    time_start = time()
    preprocessor = Preprocessor(lower_case=True)
    train_prep = [preprocessor(t) for t in train_data]
    inputs_prep = [t[0] for t in train_prep]
    targets_prep = [t[1] for t in train_prep]
    tokenizer_input = SubwordTextEncoder.build_from_corpus(
        inputs_prep,
        target_vocab_size=2 ** 15,
        max_subword_length=20)
    tokenizer_target = SubwordTextEncoder.build_from_corpus(
        targets_prep, target_vocab_size=2 ** 15,
        reserved_tokens=[preprocessor.start_token, preprocessor.end_token],
        max_subword_length=20)
    time_end = time()
    print('time tok: {}'.format((time_end - time_start)))

    vectorizer = Vectorizer(tokenizer_input, tokenizer_target)
    'vocab size input {}, target {}'.format(
        vectorizer.encoding_dim, vectorizer.decoding_dim)


    summarizer = SummarizerTransformer(num_heads=2,
                                       num_layers=1,
                                       feed_forward_dim=512,
                                       embedding_size=64,
                                       embedding_encoder_trainable=True,
                                       embedding_decoder_trainable=True,
                                       dropout_rate=0,
                                       max_prediction_len=60)
    """
    summarizer = SummarizerTransformer.load('/tmp/summarizer_20191014_195358')
#    summarizer.init_model(preprocessor, vectorizer)

    trainer = Trainer(steps_per_epoch=500,
                      batch_size=8,
                   #   embedding_path_encoder='/Users/cschaefe/datasets/glove_welt_dedup.txt',
                 #     embedding_path_decoder='/Users/cschaefe/datasets/glove_welt_dedup.txt',
              #        model_save_path='/tmp/summarizer_attention_256',
            #          tensorboard_dir='/tmp/tensorboard_attention_256',
                      steps_to_log=5)

    trainer.train(summarizer, train_data, val_data=val_data)





import logging
import sys


def get_logger(name: str) -> logging.Logger:
    logger = logging.getLogger(name)
    logger.setLevel(logging.DEBUG)
    logger.propagate = False

    if not logger.handlers:
        # stream handler ensures that logging events are passed to stdout
        ch = logging.StreamHandler(sys.stdout)
        ch.setLevel(logging.INFO)
        ch_formatter = logging.Formatter('%(message)s')
        ch.setFormatter(ch_formatter)
        logger.addHandler(ch)

    return logger



import unittest
from unittest.mock import Mock

from headliner.callbacks.validation_callback import ValidationCallback


class TestValidationCallback(unittest.TestCase):

    def test_on_epoch_end(self):
        mock_summarizer = Mock()
        mock_summarizer.new_train_step.return_value = lambda input_seq, output_seq: 0.5
        mock_scorer_a, mock_scorer_b = Mock(), Mock()
        mock_scorer_a.return_value = 1
        mock_scorer_b.return_value = 2
        val_dataset_mock = Mock()
        val_dataset_mock.take.return_value = [(1, 2), (1, 2)]
        loss_function_mock = Mock()
        validation_callback = ValidationCallback(summarizer=mock_summarizer,
                                                 val_dataset=val_dataset_mock,
                                                 loss_function=loss_function_mock,
                                                 batch_size=1)
        logs = {}
        validation_callback.on_epoch_end(0, logs=logs)
        self.assertEqual({'loss_val'}, logs.keys())
        self.assertAlmostEqual(0.5, logs['loss_val'], places=10)



import unittest
from unittest.mock import Mock

from headliner.callbacks.evaluation_callback import EvaluationCallback


class TestEvaluationCallback(unittest.TestCase):

    def test_on_epoch_end(self):
        mock_summarizer = Mock()
        mock_summarizer.predict_vectors.return_value = None
        mock_scorer_a, mock_scorer_b = Mock(), Mock()
        mock_scorer_a.return_value = 1
        mock_scorer_b.return_value = 2
        val_data = [('a', 'b'), ('c', 'd')]
        evaluation_callback = EvaluationCallback(summarizer=mock_summarizer,
                                                 scorers={'mock_score_a': mock_scorer_a,
                                                          'mock_score_b': mock_scorer_b},
                                                 val_data=val_data,
                                                 print_num_examples=0)
        logs = {}
        evaluation_callback.on_epoch_end(0, logs=logs)
        self.assertEqual({'mock_score_a', 'mock_score_b'}, logs.keys())
        self.assertAlmostEqual(1.0, logs['mock_score_a'], places=10)
        self.assertAlmostEqual(2.0, logs['mock_score_b'], places=10)



import shutil
import tempfile
import unittest
import os
from unittest.mock import Mock, call

from headliner.callbacks.model_checkpoint_callback import ModelCheckpointCallback


class TestModelCheckpointCallback(unittest.TestCase):

    def setUp(self) -> None:
        self.temp_dir = tempfile.mkdtemp(prefix='TestModelCheckpointingCallback')

    def tearDown(self) -> None:
        shutil.rmtree(self.temp_dir, ignore_errors=True)

    def test_on_epoch_end(self):
        mock_summarizer = Mock()
        model_save_path = os.path.join(self.temp_dir, 'summarizer_save')
        model_checkoint_callback = ModelCheckpointCallback(file_path=model_save_path,
                                                           summarizer=mock_summarizer,
                                                           monitor='loss_val',
                                                           mode='min')

        logs = {'loss_val': 10}
        model_checkoint_callback.on_epoch_end(0, logs=logs)
        mock_summarizer.save.assert_called_with(model_save_path)
        logs = {'loss_val': 20}
        model_checkoint_callback.on_epoch_end(0, logs=logs)
        mock_summarizer.save.assert_called_with(model_save_path)
        logs = {'loss_val': 5}
        model_checkoint_callback.on_epoch_end(0, logs=logs)
        mock_summarizer.save.assert_has_calls([call(model_save_path), call(model_save_path)])



from sklearn.model_selection import train_test_split

from headliner.model import SummarizerAttention
from headliner.model.summarizer_transformer import SummarizerTransformer
from headliner.training_runner import read_data_json

if __name__ == '__main__':
    path_to_model = '/Users/cschaefe/saved_models/att_512_7'
    summarizer = SummarizerAttention.load(path_to_model)
    data_raw = read_data_json('/Users/cschaefe/datasets/welt_dedup.json', 2000)
    train_data, val_data = train_test_split(data_raw, test_size=500, shuffle=True, random_state=42)

    for i in range(100):
        text_input, text_output = val_data[i]
        pred_text = summarizer.predict(text_input)
        print('\n\n(input) {}'.format(text_input))
        print('\n(target) {}'.format(text_output))
        print('\n(pred) {}'.format(pred_text))

    while True:
        text = input('\nEnter text: ')
        prediction_vecs = summarizer.predict_vectors(text, '')
        tokens_input = prediction_vecs['preprocessed_text'][0].split()
        print('\n')
        print(prediction_vecs['predicted_text'])
        print('\n')




from setuptools import setup, find_packages

long_description = '''
Headliner is a sequence modeling library that eases the training and 
**in particular, the deployment of custom sequence models** for both researchers and developers. 
You can very easily deploy your models in a few lines of code. It was originally 
built for our own research to generate headlines from news articles. 
That's why we chose the name, Headliner. Although this library was created internally to 
generate headlines, you can also use it for other tasks like machine translations,
text summarization and many more.

Read the documentation at: https://as-ideas.github.io/headliner/

Headliner is compatible with Python 3.6 and is distributed under the MIT license.
'''

setup(
    name='headliner',
    version='0.0.18',
    author='Christian Schäfer',
    author_email='c.schaefer.home@gmail.com',
    description='Easy training and deployment of seq2seq models.',
    long_description=long_description,
    license='MIT',
    install_requires=['scikit-learn', 'nltk', 'pyyaml'],
    extras_require={
        'tests': ['pytest', 'pytest-cov', 'codecov', 'tensorflow~=2.0.0'],
        'docs': ['mkdocs', 'mkdocs-material'],
        'dev': ['bumpversion']
    },
    classifiers=[
        'Development Status :: 4 - Beta',
        'Intended Audience :: Developers',
        'Intended Audience :: Education',
        'Intended Audience :: Science/Research',
        'License :: OSI Approved :: MIT License',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.6',
        'Topic :: Software Development :: Libraries',
        'Topic :: Software Development :: Libraries :: Python Modules',
    ],
    packages=find_packages(exclude=('tests',)),
)